/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdlib.h>
#include <string.h>
#include <memory.h>

#include <mplib1/mplib1_config.h>

#include <mplib1/dl_list.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

dl_Node_t *
dl_Init_Node( dl_Node_t *node, const char *name, const void *item )
{
    node->ln_Succ = NULL;
    node->ln_Pred = NULL;
    node->ln_Name = (char *)(size_t)name;
    node->ln_Item = (void *)(size_t)item;
    node->ln_List = NULL;
    return(node);
}

dl_List_t *
dl_Init_List( dl_List_t *list, int flags )
{
    if (list)
    {
	(void)dl_Init_Node( &list->ln_Head, NULL, NULL );
	(void)dl_Init_Node( &list->ln_Tail, NULL, NULL );
	list->ln_Head.ln_List = list;
	list->ln_Tail.ln_List = list;
	list->ln_Flags = flags;
	list->ln_Head.ln_Pred = NULL;
	list->ln_Head.ln_Succ = &list->ln_Tail;
	list->ln_Tail.ln_Pred = &list->ln_Head;
	list->ln_Tail.ln_Succ = NULL;
    }
    return(list);
}


dl_Node_t *
dl_Remove_Node( dl_Node_t *node1 )
{
    if (node1)
    {
	if (node1->ln_Succ && node1->ln_Succ->ln_Pred == node1)
	    node1->ln_Succ->ln_Pred = node1->ln_Pred;
	if (node1->ln_Pred && node1->ln_Pred->ln_Succ == node1)
	    node1->ln_Pred->ln_Succ = node1->ln_Succ;
	node1->ln_Succ = NULL;
	node1->ln_Pred = NULL;
    }
    return(node1);
}


dl_Node_t *
dl_Remove_Head( dl_List_t *list )
{
    dl_Node_t *node=NULL;
    if (list->ln_Head.ln_Succ != &list->ln_Tail)
	node = dl_Remove_Node(list->ln_Head.ln_Succ);
    return(node);
}


dl_Node_t *
dl_Remove_Tail( dl_List_t *list )
{
    dl_Node_t *node=NULL;
    if (list->ln_Tail.ln_Pred != &list->ln_Head)
	node = dl_Remove_Node(list->ln_Tail.ln_Pred);
    return(node);
}


dl_Node_t *
dl_Add_Node_After( dl_Node_t *node1, dl_Node_t *node2 )
{
    node2->ln_Succ = node1->ln_Succ;
    node2->ln_Pred = node1;
    node1->ln_Succ->ln_Pred = node2;
    node1->ln_Succ = node2;
    return(node2);
}


dl_Node_t *
dl_Add_Node_Before( dl_Node_t *node1, dl_Node_t *node2 )
{
    node2->ln_Succ = node1;
    node2->ln_Pred = node1->ln_Pred;
    node1->ln_Pred->ln_Succ = node2;
    node1->ln_Pred = node2;
    return(node2);
}


dl_Node_t *
dl_Add_Head( dl_List_t *list, dl_Node_t *node )
{
    (void)dl_Remove_Node(node);

    (void)dl_Add_Node_After( &list->ln_Head, node );
    node->ln_List = list;

    return(node);
}


dl_Node_t *
dl_Add_Tail( dl_List_t *list, dl_Node_t *node )
{
    (void)dl_Remove_Node(node);

    (void)dl_Add_Node_Before( &list->ln_Tail, node );
    node->ln_List = list;

    return(node);
}


void *
dl_Find_Item_By_Name( dl_List_t *list, const char *name )
{
    void *item=NULL;
    dl_Node_t *node;

    if (name && *name)
    {
	node = list->ln_Head.ln_Succ;
	/* walk the list until we find the matching entry */
	while (node->ln_Succ && item == NULL)
	{
	    if (   node->ln_Name
		&& (strcmp(name,node->ln_Name)==0)
	       )
		item = (void *)node->ln_Item;
	    node = node->ln_Succ;
	};
    }
	    
    return(item);
}


void *
dl_Find_Next_Item_By_Name( dl_Node_t *node, const char *name )
{
    void *item=NULL;

    if (name && *name)
    {
	node = node->ln_Succ;
	/* walk the list until we find the matching entry */
	while (node->ln_Succ && item == NULL)
	{
	    if (   node->ln_Name
		&& (strcmp(name,node->ln_Name)==0)
	       )
		item = (void *)node->ln_Item;
	    node = node->ln_Succ;
	};
    }
	    
    return(item);
}


dl_Node_t *
dl_Find_Node_By_Item( dl_List_t *list, const void *item )
{
    dl_Node_t *node,*rnode=NULL;

    node = list->ln_Head.ln_Succ;
    while (node->ln_Succ && rnode == NULL)
    {
	if (node->ln_Item == item)
	    rnode = node;
	node = node->ln_Succ;
    };
    return(rnode);
}


void *
dl_Remove_Head_Item( dl_List_t *list )
{
    dl_Node_t *node=NULL;
    if (list->ln_Head.ln_Succ != &list->ln_Tail)
	node = dl_Remove_Node(list->ln_Head.ln_Succ);
    return( (void *)((node)?node->ln_Item:NULL) );
}


void *
dl_Remove_Tail_Item( dl_List_t *list )
{
    dl_Node_t *node=NULL;
    if (list->ln_Tail.ln_Pred != &list->ln_Head)
	node = dl_Remove_Node(list->ln_Tail.ln_Pred);
    return( (void *)((node)?node->ln_Item:NULL) );
}


int
dl_Walk_List( dl_List_t *l_ptr,
	    dl_Walk_List_t disp,
	    void *param )
{
    dl_Node_t *n_ptr,*nn_ptr;

    n_ptr = l_ptr->ln_Head.ln_Succ;

    while (n_ptr->ln_Succ)
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = n_ptr->ln_Succ;
	(*disp)( (void *)n_ptr->ln_Item, param );
	if (n_ptr == nn_ptr)
	    break;
	n_ptr = nn_ptr;
    };
    return(0);
}


int
dl_Walk_List_Name( dl_List_t *l_ptr,
		const char *name,
		dl_Walk_List_Name_t disp,
		void *param )
{
    dl_Node_t *n_ptr,*nn_ptr;
    int rv=0;

    n_ptr = l_ptr->ln_Head.ln_Succ;

    while (n_ptr->ln_Succ && rv==0 )
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = n_ptr->ln_Succ;
	if (   n_ptr->ln_Name
	    && (strcmp(name,n_ptr->ln_Name)==0)
	   )
	    rv = (*disp)( (void *)n_ptr->ln_Item, param );
	if (n_ptr == nn_ptr)
	    break;
	n_ptr = nn_ptr;
    };
    return(rv);
}


int
dl_Transfer_Lists( dl_List_t *fl_ptr, dl_List_t *tl_ptr )
{
    dl_Node_t *n_ptr, *np2=NULL;
    int rv=0;
    while( (n_ptr=dl_Remove_Head( fl_ptr )) )
    {
	if (np2==n_ptr)
	    break;	/* Problem with list */
	np2=n_ptr;
	(void)dl_Add_Tail( tl_ptr, n_ptr );
	rv++;
    }
    return(rv);
}


int
dl_Any_In_List( dl_List_t *list )
{
    return( (list->ln_Head.ln_Succ->ln_Succ != NULL) );
}


int
dl_Walk_List_By_Node( dl_List_t *l_ptr,
		    dl_Walk_List_Node_t disp,
		    void *param )
{
    dl_Node_t *n_ptr,*nn_ptr;
    int rv=0;

    n_ptr = l_ptr->ln_Head.ln_Succ;

    while (n_ptr->ln_Succ && rv==0)
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = n_ptr->ln_Succ;
	rv = (*disp)( n_ptr, param );
	n_ptr = nn_ptr;
    };
    return(rv);
}


int
dl_Free_List_Contents( dl_List_t *fl_ptr )
{
    dl_Node_t *n_ptr, *np2=NULL;
    int rv=0;
    while( (n_ptr=dl_Remove_Head( fl_ptr )) )
    {
	if (np2==n_ptr)
	    break;	/* Problem with list */
	if( n_ptr->ln_Item )
	    free( (void *)n_ptr->ln_Item );
	else
	    free( n_ptr );
	np2=n_ptr;
	rv++;
    }
    return(rv);
}


int
dl_Walk_List2( dl_List_t *l_ptr,
	    dl_Walk_List2_t disp,
	    void *param )
{
    dl_Node_t *n_ptr,*nn_ptr;
    int rv = 0;

    n_ptr = l_ptr->ln_Head.ln_Succ;

    while (n_ptr->ln_Succ && rv==0 )
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = n_ptr->ln_Succ;
	rv = (*disp)( (void *)n_ptr->ln_Item, param );
	if (n_ptr==nn_ptr)
	    break;
	n_ptr = nn_ptr;
    };
    return(rv);
}


static void
dl_Free_Node_private( dl_Node_t *nptr, int *flags )
{
    if (   (*flags & DL_FREE_NODES )
	|| (*flags & DL_REMOVE_NODES)
       )
	dl_Remove_Node( nptr );

    if (   (*flags & DL_FREE_NAMES)
	&& nptr->ln_Name
       )
    {
	free( (void *)nptr->ln_Name );
	nptr->ln_Name = NULL;
    }
    if (   (*flags & DL_FREE_ITEMS)
	&& nptr->ln_Item
       )
    {
	void *vp;	/* just in case the node is in the item */
	vp = nptr->ln_Item;
	nptr->ln_Item = NULL;
	free( vp );
    }
    if ( *flags & DL_FREE_NODES )
    {
	free( nptr );
    }
    return;
}

void
dl_Free_Node( dl_Node_t *nptr, int flags )
{
    dl_Free_Node_private( nptr, &flags );
    return;
}

void
dl_Free_List( dl_List_t *fl_ptr, int flags )
{
    dl_Walk_List( fl_ptr, (dl_Walk_List_t)dl_Free_Node_private, (void *)&flags );
    return;
}

/* -- End of File -- */
