/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#include <mplib1/mplib1_config.h>
#include <mplib1/stopwatch.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static int
get_time( struct timeval *tvp )
{
    int rv;
#ifdef H_GETTIMEOFDAY_2_NULL
    rv = gettimeofday( tvp, NULL );
#else
#ifdef H_GETTIMEOFDAY_2_PARAMS
    struct timezone my_tz;

    my_tz.tz_minuteswest = 0;
    my_tz.tz_dsttime = 0;

    rv = gettimeofday( tvp, &my_tz );
#else
    rv = gettimeofday( tvp );
#endif
#endif
    return(rv);
}

int
init_stopwatch( struct stopwatch *stp )
{
    if (stp)
	memset( stp, '\0', sizeof(struct stopwatch) );
    return(0);
}


int
start_stopwatch( struct stopwatch *stp )
{
    int rv;
    if (stp)
    {
	rv = get_time( &stp->start_time );
    }else
    {
	rv= -1;
	errno = ENOENT;
    }
    return(rv);
}

int
stop_stopwatch( struct stopwatch *stp )
{
    int rv;
    struct timeval end_time;
    if (stp && stp->start_time.tv_sec)
    {
	rv = get_time( &end_time );
	accum_stop_timevals( stp, &end_time );
    }else
    {
	rv= -1;
	errno = ENOENT;
    }
    return(rv);
}

int
accrete_stopwatch( struct stopwatch *stp_into, struct stopwatch *stp_from )
{
    add_timevals( &stp_into->cumulative_time, &stp_from->cumulative_time );
    return(0);
}

/* -- End of File -- */
