/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <mplib1/mplib1_config.h>

#include "../include/bpo_internal.h"
#include <mplib1/bpo.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

bpo_Node_t *
bpo_Init_Node( bpo_Node_t *node, const char *name, void *item )
{
    void *seg_base;

    seg_base = Get_SODB_Base( node );
    if (!seg_base)
	return(NULL);
    /* Initial requiremnt, name, item & list must be null of in same SODB as Node */
    if (   (name==NULL || Get_SODB_Base( name ) == seg_base )
	&& (item==NULL || Get_SODB_Base( item ) == seg_base )
       )
    {
	node->lno_me = GOFF(seg_base,node);
	node->lno_Succ = 0;
	node->lno_Pred = 0;
	if (name)
	    node->lno_Name = GOFF( seg_base, name );
	else
	    node->lno_Name = 0;
	if (item)
	    node->lno_Item = GOFF( seg_base, item );
	else
	    node->lno_Item = 0;
	node->lno_List = 0;
    }else
	return(NULL);
    return(node);
}


bpo_List_t *
bpo_Init_List( bpo_List_t *list, int flags )
{
    char *seg_base;

    if (list==NULL)
	return(NULL);

    seg_base = Get_SODB_Base( list );
    if (!seg_base)
	return(NULL);

    bpo_Init_pid_lock( &list->lno_Pid_Lock, NULL );
    bpo_pid_Lock( &list->lno_Pid_Lock );
    list->lno_me = GOFF( seg_base, list );
    list->lno_Head.lno_me = GOFF( seg_base, &list->lno_Head );
    list->lno_Tail.lno_me = GOFF( seg_base, &list->lno_Tail );
    list->lno_Head.lno_Pred = 0;
    list->lno_Head.lno_Succ = list->lno_Tail.lno_me;
    list->lno_Tail.lno_Pred = list->lno_Head.lno_me;
    list->lno_Tail.lno_Succ = 0;
    list->lno_Head.lno_List = GOFF( seg_base, list );
    list->lno_Tail.lno_List = GOFF( seg_base, list );
    list->lno_Flags = flags;

    list->lno_count=0;
    list->lno_serial_nbr = 1L;
    bpo_pid_Unlock( &list->lno_Pid_Lock );

    return(list);
}


bpo_Node_t *
bpo_Remove_Node( bpo_Node_t *node1 )
{
    bpo_List_t *lp;
    bpo_Node_t *nnode,*pnode;

    if (Get_SODB_Base( node1 )==NULL)
	return(NULL);

    if (node1 && node1->lno_Succ && node1->lno_Pred && node1->lno_List)
    {
	lp = GRPTR( node1, node1->lno_List );
	if (lp && lp->lno_me == node1->lno_List )
	{
	    bpo_pid_Lock( &lp->lno_Pid_Lock );
	    nnode = GNNODE( node1, node1->lno_Succ );
	    pnode = GNNODE( node1, node1->lno_Pred );
	    if (nnode && nnode->lno_Pred == node1->lno_me)
		nnode->lno_Pred = node1->lno_Pred;
	    if (pnode && pnode->lno_Succ == node1->lno_me)
		pnode->lno_Succ = node1->lno_Succ;
	    node1->lno_Succ = 0;
	    node1->lno_Pred = 0;
	    node1->lno_List = 0;
	    lp->lno_count--;
	    lp->lno_serial_nbr++;
	    bpo_pid_Unlock( &lp->lno_Pid_Lock );
	}

    }
    return(node1);
}


bpo_Node_t *
bpo_Remove_Head( bpo_List_t *list )
{
    bpo_Node_t *node=NULL;
    if (list->lno_Head.lno_Succ != list->lno_Tail.lno_me)
	node = bpo_Remove_Node( GNNODE( &list->lno_Head, list->lno_Head.lno_Succ));
    return(node);
}




bpo_Node_t *
bpo_Remove_Tail( bpo_List_t *list )
{
    bpo_Node_t *node=NULL;
    if (list->lno_Tail.lno_Pred != list->lno_Head.lno_me)
	node = bpo_Remove_Node( GNNODE( &list->lno_Head, list->lno_Tail.lno_Pred));
    return(node);
}


bpo_Node_t *
bpo_Add_Node_After( bpo_Node_t *node1, bpo_Node_t *node2 )
{
    bpo_List_t *lp;
    if (Get_SODB_Base( node1 ) != Get_SODB_Base( node2 ))
	return(NULL);
    lp = GRPTR( node1, node1->lno_List );
    if (lp && lp->lno_me==node1->lno_List)
    {
	bpo_pid_Lock( &lp->lno_Pid_Lock );
	if (node1->lno_Succ && node2->lno_Succ == (off_t)0)
	{
	    node2->lno_Succ = node1->lno_Succ;
	    node2->lno_Pred = node1->lno_me;
	    GNNODE( node1, node1->lno_Succ )->lno_Pred = node2->lno_me;
	    node1->lno_Succ = node2->lno_me;
	    node2->lno_List = node1->lno_List;
	    lp->lno_count++;
	    lp->lno_serial_nbr++;
	}
	bpo_pid_Unlock( &lp->lno_Pid_Lock );
    }
    return(node2);
}


bpo_Node_t *
bpo_Add_Node_Before( bpo_Node_t *node1, bpo_Node_t *node2 )
{
    bpo_List_t *lp;
    if (Get_SODB_Base( node1 ) != Get_SODB_Base( node2 ))
	return(NULL);
    lp = GRPTR( node1, node1->lno_List );
    if (lp && lp->lno_me==node1->lno_List)
    {
	bpo_pid_Lock( &lp->lno_Pid_Lock );
	if (node1->lno_Pred && node2->lno_Pred==(off_t)0)
	{
	    node2->lno_Succ = node1->lno_me;
	    node2->lno_Pred = node1->lno_Pred;
	    GNNODE( node1, node1->lno_Pred)->lno_Succ = node2->lno_me;
	    node1->lno_Pred = node2->lno_me;
	    node2->lno_List = node1->lno_List;
	    lp->lno_count++;
	    lp->lno_serial_nbr++;
	}
	bpo_pid_Unlock( &lp->lno_Pid_Lock );
    }
    return(node2);
}


bpo_Node_t *
bpo_Add_Head( bpo_List_t *list, bpo_Node_t *node )
{
    bpo_Remove_Node(node);

    bpo_Add_Node_After( &list->lno_Head, node );

    return(node);
}



bpo_Node_t *
bpo_Add_Tail( bpo_List_t *list, bpo_Node_t *node )
{
    if (node->lno_Succ || node->lno_Pred || node->lno_List)
	bpo_Remove_Node(node);

    bpo_Add_Node_Before( &list->lno_Tail, node );

    return(node);
}


void *
bpo_Find_Item_By_Name( bpo_List_t *list, const char *name )
{
    void *item=NULL;
    bpo_Node_t *node;
    void *seg_base;

    seg_base = Get_SODB_Base( list );
    if (seg_base==NULL)
	return(NULL);

    bpo_pid_Lock( &list->lno_Pid_Lock );
    node = GNNODE( &list->lno_Head, list->lno_Head.lno_Succ );
    /* walk the list until we find the matching entry */
    while (node->lno_Succ)
    {
	if (strcmp(name,(const char *)(GPTR( seg_base, node->lno_Name)))==0)
	{
	    bpo_pid_Unlock( &list->lno_Pid_Lock );
	    return( GPTR( seg_base, node->lno_Item ) );
	}
	node = GNNODE( node, node->lno_Succ);
    };
    bpo_pid_Unlock( &list->lno_Pid_Lock );

    return(item);
}


void *
bpo_Find_Next_Item_By_Name( bpo_Node_t *node, const char *name )
{
    void *item=NULL;
    void *seg_base;
    bpo_List_t *list;

    seg_base = Get_SODB_Base( node );
    if (seg_base==NULL)
	return(NULL);

    list = GPTR( seg_base, node->lno_List );

    bpo_pid_Lock( &list->lno_Pid_Lock );
    node = GNNODE( node, node->lno_Succ);
    /* walk the list until we find the matching entry */
    while (node && node->lno_Succ && item == NULL)
    {
	if (strcmp(name,GPTR( seg_base, node->lno_Name))==0)
	{
	    bpo_pid_Unlock( &list->lno_Pid_Lock );
	    return( GPTR( seg_base, node->lno_Item ) );
	}
	node = GNNODE( node, node->lno_Succ);
    };
    bpo_pid_Unlock( &list->lno_Pid_Lock );

    return(item);
}

bpo_Node_t *
bpo_Find_Node_By_Item( bpo_List_t *list, void *item )
{
    bpo_Node_t *node,*rnode=NULL;


    bpo_pid_Lock( &list->lno_Pid_Lock );
    node = GNNODE( &list->lno_Head, list->lno_Head.lno_Succ);
    while (node->lno_Succ && rnode == NULL)
    {
	if (GRPTR( node, node->lno_Item ) == item)
	{
	    bpo_pid_Unlock( &list->lno_Pid_Lock );
	    return( node );
	}
	node = GNNODE( node, node->lno_Succ );
    };
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rnode);
}

void *
bpo_Find_Item_From_Node( bpo_Node_t *node )
{
    return( GRPTR( node, node->lno_Item ) );
}

void *
bpo_Remove_Head_Item( bpo_List_t *list )
{
    bpo_Node_t *node=NULL;
    if (list->lno_Head.lno_Succ != list->lno_Tail.lno_me)
	node = bpo_Remove_Node( GNNODE( &list->lno_Head, list->lno_Head.lno_Succ));
    return( (void *)((node)?GRPTR( node, node->lno_Item):NULL) );
}


void *
bpo_Remove_Tail_Item( bpo_List_t *list )
{
    bpo_Node_t *node=NULL;
    if (list->lno_Tail.lno_Pred != list->lno_Head.lno_me)
	node = bpo_Remove_Node( GNNODE( &list->lno_Tail, list->lno_Tail.lno_Pred ));
    return( (void *)((node)?GRPTR( node, node->lno_Item):NULL) );
}

void
bpo_Walk_List( bpo_List_t *l_ptr,
		bpo_Walk_List_t disp,
		void *param )
{
    bpo_Node_t *n_ptr,*nn_ptr;
    void *seg_base;

    seg_base = Get_SODB_Base( l_ptr );

    bpo_pid_Lock( &l_ptr->lno_Pid_Lock );
    n_ptr = GANNODE( seg_base, l_ptr->lno_Head.lno_Succ );

    while (n_ptr->lno_Succ)
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = GANNODE( seg_base, n_ptr->lno_Succ);
	(*disp)( GRPTR( n_ptr, n_ptr->lno_Item ), param );
	if (n_ptr == nn_ptr)
	    break;
	n_ptr = nn_ptr;
    };
    bpo_pid_Unlock( &l_ptr->lno_Pid_Lock );
    return;
}

int
bpo_Walk_List2( bpo_List_t *l_ptr,
		bpo_Walk_List2_t disp,
		void *param )
{
    bpo_Node_t *n_ptr,*nn_ptr;
    int rv = 0;
    char *seg_base;

    seg_base = Get_SODB_Base( l_ptr );

    bpo_pid_Lock( &l_ptr->lno_Pid_Lock );
    n_ptr = GANNODE( seg_base, l_ptr->lno_Head.lno_Succ );

    while (n_ptr->lno_Succ && rv==0 )
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = GANNODE( seg_base, n_ptr->lno_Succ);
	rv = (*disp)( GRPTR( n_ptr, n_ptr->lno_Item ), param );
	if (n_ptr == nn_ptr)
	    break;
	n_ptr = nn_ptr;
    };
    bpo_pid_Unlock( &l_ptr->lno_Pid_Lock );
    return(rv);
}

int
bpo_Walk_List_N( bpo_List_t *l_ptr,
		bpo_Walk_List_N_t disp,
		void *param )
{
    bpo_Node_t *n_ptr,*nn_ptr;
    int rv = 0;
    char *seg_base;

    seg_base = Get_SODB_Base( l_ptr );

    bpo_pid_Lock( &l_ptr->lno_Pid_Lock );
    n_ptr = GANNODE( seg_base, l_ptr->lno_Head.lno_Succ );

    while (n_ptr->lno_Succ && rv==0 )
    {
	/* save the next pointer just in case they mash the list */
	nn_ptr = GANNODE( seg_base, n_ptr->lno_Succ);
	rv = (*disp)( n_ptr, param );
	if (n_ptr == nn_ptr)
	    break;
	n_ptr = nn_ptr;
    };
    bpo_pid_Unlock( &l_ptr->lno_Pid_Lock );
    return(rv);
}

bpo_Node_t *
bpo_Init_Node_Raw( bpo_Node_t *node,
		    const void *base,
		    const char *name,
		    const void *item )
{
    node->lno_me = GOFF(base,node);
    node->lno_Succ = 0;
    node->lno_Pred = 0;
    if (name)
	node->lno_Name = GOFF( base, name );
    else
	node->lno_Name = 0;
    if (item)
	node->lno_Item = GOFF( base, item );
    else
	node->lno_Item = 0;
    node->lno_List = 0;
    return(node);
}

bpo_List_t *
bpo_Init_List_Raw( bpo_List_t *list, const void *base, int flags )
{
    if (list==NULL)
	return(NULL);

    bpo_Init_pid_lock( &list->lno_Pid_Lock, NULL );
    bpo_pid_Lock( &list->lno_Pid_Lock );
    list->lno_me = GOFF( base, list );
    list->lno_Head.lno_me = GOFF( base, &list->lno_Head );
    list->lno_Tail.lno_me = GOFF( base, &list->lno_Tail );
    list->lno_Head.lno_Pred = 0;
    list->lno_Head.lno_Succ = list->lno_Tail.lno_me;
    list->lno_Tail.lno_Pred = list->lno_Head.lno_me;
    list->lno_Tail.lno_Succ = 0;
    list->lno_Head.lno_List = GOFF( base, list );
    list->lno_Tail.lno_List = GOFF( base, list );
    list->lno_Flags = flags;
    list->lno_count = 0;
    list->lno_serial_nbr = 1L;
    bpo_pid_Unlock( &list->lno_Pid_Lock );

    return(list);
}

bpo_Node_t *
bpo_Add_Head_Raw( bpo_List_t *list, const void *base, bpo_Node_t *node )
{
    bpo_pid_Lock( &list->lno_Pid_Lock );
    node->lno_Succ = list->lno_Head.lno_Succ;
    node->lno_Pred = list->lno_Head.lno_me;
    GNNODE( &list->lno_Head, list->lno_Head.lno_Succ )->lno_Pred = node->lno_me;
    list->lno_Head.lno_Succ = node->lno_me;

    node->lno_List = list->lno_me;
    list->lno_count++;
    list->lno_serial_nbr++;
    bpo_pid_Unlock( &list->lno_Pid_Lock );

    return(node);
}

bpo_Node_t *
bpo_Add_Tail_Raw( bpo_List_t *list, const void *base, bpo_Node_t *node )
{
    bpo_pid_Lock( &list->lno_Pid_Lock );
    node->lno_Succ = list->lno_Tail.lno_me;
    node->lno_Pred = list->lno_Tail.lno_Pred;
    GNNODE( &list->lno_Tail, list->lno_Tail.lno_Pred)->lno_Succ = node->lno_me;
    list->lno_Tail.lno_Pred = node->lno_me;

    node->lno_List = list->lno_me;
    list->lno_count++;
    list->lno_serial_nbr++;
    bpo_pid_Unlock( &list->lno_Pid_Lock );

    return(node);
}

int
bpo_Work_List( bpo_List_t *list,
		int (* work)( bpo_List_t *, void * ),
		void *param )
{
    int rv;
    bpo_pid_Lock( &list->lno_Pid_Lock );
    rv = (*work)( list, param );
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rv);
}

int
bpo_Work_List2( bpo_List_t *list,
		int (* work)( bpo_List_t *, void *, void * ),
		void *param1, void *param2 )
{
    int rv;
    bpo_pid_Lock( &list->lno_Pid_Lock );
    rv = (*work)( list, param1, param2 );
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rv);
}

int
bpo_May_Work_List( bpo_List_t *list,
		    unsigned long *last_ser_nbr,
		    int (* work)( bpo_List_t *, void * ),
		    void *param )
{
    int rv=0;
    bpo_pid_Lock( &list->lno_Pid_Lock );
    if (list->lno_serial_nbr > *last_ser_nbr)
    {
	rv = (*work)( list, param );
	*last_ser_nbr = list->lno_serial_nbr;
    }
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rv);
}

int
bpo_May_Work_List2( bpo_List_t *list,
		    unsigned long *last_ser_nbr,
		    int (* work)( bpo_List_t *, void *, void * ),
		    void *param1, void *param2 )
{
    int rv=0;
    bpo_pid_Lock( &list->lno_Pid_Lock );
    if (list->lno_serial_nbr > *last_ser_nbr)
    {
	rv = (*work)( list, param1, param2 );
	*last_ser_nbr = list->lno_serial_nbr;
    }
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rv);
}

int
bpo_Transfer_Lists( bpo_List_t *fl_ptr, bpo_List_t *tl_ptr )
{
    bpo_Node_t *n_ptr, *np2=NULL;
    int rv=0;
    while( (n_ptr=bpo_Remove_Head( fl_ptr )) )
    {
	if (np2==n_ptr)
	    break;	/* Problem with list */
	np2=n_ptr;
	bpo_Add_Tail( tl_ptr, n_ptr );
	rv++;
    }
    return(rv);
}

unsigned long
bpo_Touch_List( bpo_List_t *list )
{
    unsigned long rv;
    bpo_pid_Lock( &list->lno_Pid_Lock );
    list->lno_serial_nbr++;
    rv = list->lno_serial_nbr;
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rv);
}


static void
remove_and_free( bpo_List_t *blp )
{
    void *vp;
    bpo_Node_t *bnp,*bnp2=NULL;

    while( (bnp=bpo_Remove_Head( blp )) )
    {
	if (bnp==bnp2)
	    break;	/* Problem with list */
	bnp2 = bnp;
	vp = GRPTR( bnp, bnp->lno_Item );
	if (vp)
	    shfree(vp);
    };
    blp->lno_count=0;
    return;
}

int
bpo_free_list_contents( bpo_List_t *blp )
{
    bpo_Work_List( blp,
		(int(*)(bpo_List_t *, void *))remove_and_free,
		NULL );
    return(0);
}


static void
internal_bpo_Free_Node( bpo_Node_t *bnp, int *flags )
{
    if (   (*flags & BPO_REMOVE_NODES)
	|| (*flags & BPO_FREE_NODES)
       )
	bpo_Remove_Node( bnp );
    if (   (*flags & BPO_FREE_NAMES)
	&& bnp->lno_Name
       )
    {
	shfree( Get_SODB_Pointer( bnp, bnp->lno_Name ) );
	bnp->lno_Name = (off_t)0;
    }
    if (   (*flags & BPO_FREE_ITEMS)
	&& bnp->lno_Item
       )
    {
	shfree( Get_SODB_Pointer( bnp, bnp->lno_Item ) );
	bnp->lno_Item = (off_t)0;
    }
    if ( *flags & BPO_FREE_NODES )
    {
	shfree( bnp );
    }
    return;
}

void
bpo_Free_Node( bpo_Node_t *bnp, int flags )
{
    internal_bpo_Free_Node( bnp, &flags );
    return;
}

void
bpo_Free_List( bpo_List_t *blp, int flags )
{
    bpo_Walk_List( blp,
		(bpo_Walk_List_t)internal_bpo_Free_Node,
		&flags );
    return;
}


int
bpo_Any_In_List( bpo_List_t *list )
{
    int rv;
    bpo_pid_Lock( &list->lno_Pid_Lock );
    rv=list->lno_count;
    bpo_pid_Unlock( &list->lno_Pid_Lock );
    return(rv);
}


/* -- End of File -- */
