/*
*******************************************************************************
*  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <signal.h>
#include <dirent.h>

#include <mplib1/mplib1_config.h>
#include "../include/bpo_internal.h"
#include <mplib1/bpo.h>

#include <mplib1/cfg_file.h>
#include <mplib1/safe_string.h>
#include <mplib1/pid_check.h>
#include <mplib1/fprintfile.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static int
check_pipe( char *name )
{
    /* Is this one a pipe, and is there a pid on the end, and is it alive */
    char *cp;
    int rv = 0;
    struct stat tstat;
    pid_t pid;

    /* is it a pipe ? */
    if (   stat(name, &tstat)==0
	&& S_ISFIFO(tstat.st_mode)
       )
    {
	cp = name + Sstrlen(name) - 1;
	if (isdigit(*cp))
	{
	    /* looks like a pid, so scan backwards to beginning */
	    while (isdigit(*(cp-1)))
		cp--;
	    pid = atoi(cp);
	    if (pid && is_pid_dead(pid)!=0)
	    {
		/* it is a pid and the pid is not there */
		unlink( name );
		rv = 1;
	    }
	}
    }
    return(rv);
}

int
scan_dead_pipes( const char *dir_name )
{
    DIR *dirp;
    struct dirent *direntp;
    char *olddir;
    int rv = 0;

    olddir = getcwd( NULL, 200 );
    if (chdir(dir_name)==0)
    {
	dirp = opendir( dir_name );
	while ((direntp=readdir(dirp))!=NULL)
	{
	    check_pipe( direntp->d_name );
	};
	closedir(dirp);
	chdir(olddir);
    }
    free(olddir);
    return(rv);
}

/* -- End of File -- */
