#ifndef MPLIB1_BPO
#define MPLIB1_BPO
/*
*******************************************************************************
**  Copyright (c) 1996 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

#ifndef MPLIB1_CONFIG
#include <mplib1/mplib1_config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>



/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

#define GOFF(base,obj) ((off_t)((const char *)(obj) - (const char *)(base)))
#define GPTR(base,off) ((void *)((char *)(base) + off))
#define GRPTR(obj,off) (void *)((off)?(void *)((char *)(obj) - \
			*((off_t *)(obj)) + (off_t)(off)):NULL)

#ifndef BIT
#define BIT(x)	(1<<(x))
#endif

#define SODB_USE_SHALLOC	BIT(0)
#define SODB_USE_RESOURCES	BIT(0)
#define SODB_USE_LIMIT		BIT(1)
#define SODB_ADD_HEADER		BIT(2)

#define SODB_BASE_NAME	"SODB Base"

#define CHECK_ALL_RESOURCES 0

#ifndef SODB_MMAP_PRIV
#define SODB_MMAP_PRIV	0600
#endif
#ifndef SODB_IPC_PRIV
#define SODB_IPC_PRIV	0600
#endif

#define SODB_RES_NO_ALLOC_NM	BIT(0)

#ifndef SODB_READ_CONFIG
#define SODB_READ_CONFIG	BIT(0)
#endif

typedef int (*Validate_Resource_t)( const char * res_name,
				    off_t res_off,
				    void *param1 );

typedef void (*Fetch_SODB_Resource_t)( void *);



enum bpo_lock_mode
{
BPO_L_READ,
BPO_L_WRITE
};


#define BPO_PID_F_DEBUG	BIT(0)


#define SHM_INHERIT_GROUP	BIT(0)
#define SHM_INHERIT_PROGRAM	BIT(1)
#define SHM_AUTOMAGIC_PIPE	BIT(2)

#define SHM_CHECK_ALL_PROCESSES		1
#define SHM_CHECK_THIS_PROCESS		2
#define SHM_CHECK_DEAD_PROCESSES	3

/* Return values to the validate processes inform call */
#define SHM_LEAVE_PROC	0
#define SHM_DELETE_PROC	1

#ifndef PID_STR_LEN
#define PID_STR_LEN	12
#endif

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

#define Q_NM_UNKNOWN	0
#define Q_NM_NONE	1
#define	Q_NM_PIPE	2

#ifndef BIT
#define BIT(x)	(1<<(x))
#endif

#define Q_SOLE_READER	BIT(0)
#define Q_MULTI_READ	BIT(1)
#define Q_WRITER	BIT(2)
#define Q_PERSISTENT	BIT(4)

#define Q_DELETED	BIT(8)

#define Q_DEBUG		BIT(15)

#define Q_CHECK_ALL_QUEUES		1
#define Q_CHECK_THIS_QUEUE		2
#define Q_CHECK_DEAD_QUEUES	3

/* Return values to the validate processes inform call */
#define Q_LEAVE_QUEUE	0
#define Q_DELETE_QUEUE	1


/* ------------------------------------------------------------------
    bpo lock structures
   ------------------------------------------------------------------ */

struct bpo_thread_lock
{
pthread_mutex_t		mp;
pthread_mutexattr_t	mattr;
};

typedef struct bpo_thread_lock b_lock_t;

struct use_count_lock
{
b_lock_t	ucl_b_lock;
int	ucl_count;
};

struct bpo_fl_lock
{
b_lock_t	bpo_fl_lock;
int	r_cnt;
int	w_cnt;
};

struct bpo_pid_lock
{
b_lock_t	bpo_pid_lock;
pid_t	lock_pid;
int	pid_cnt;
b_lock_t	bpo_unlock;
pid_t	unlock_pid;
int	flags;
struct	timeval	lock_time;
struct	timeval	lock_duration;
};

/* ------------------------------------------------------------------
    bpo_list structures

    These structures should never be directly modified.

    The structures are included here so they can be incorporated
    into larger application structures

   ------------------------------------------------------------------ */

#define BPO_NODE_TYPE	bpo_Node_s
#define BPO_LIST_TYPE	bpo_List_s

struct BPO_NODE_TYPE
{
off_t	lno_me;
off_t	lno_Succ;
off_t	lno_Pred;
off_t	lno_Name;
off_t	lno_Item;
off_t	lno_List;
int	lno_Flags;
};

struct BPO_LIST_TYPE
{
off_t	lno_me;
struct	BPO_NODE_TYPE	lno_Head;
struct	BPO_NODE_TYPE	lno_Tail;
int			lno_Flags;
int			lno_count;
unsigned	long	lno_serial_nbr;


#ifndef LN_IGNORECASE
#define LN_IGNORECASE   BIT(0)
#endif

struct	bpo_pid_lock	lno_Pid_Lock;
};


typedef struct BPO_NODE_TYPE		bpo_Node_t;
typedef struct BPO_LIST_TYPE		bpo_List_t;

#define BPO_NODE_SIZE		sizeof(bpo_Node_t)
#define BPO_LIST_SIZE		sizeof(bpo_List_t)

#define BPO_REMOVE_NODES	BIT(0)
#define BPO_FREE_NODES		BIT(1)
#define BPO_FREE_ITEMS		BIT(2)
#define BPO_FREE_NAMES		BIT(3)

typedef void (*bpo_Walk_List_t) ( void *, void * );
typedef int (*bpo_Walk_List2_t) ( void *, void * );
typedef int (*bpo_Walk_List_N_t) ( bpo_Node_t *, void * );
typedef int (*bpo_Work_List_t) ( bpo_List_t *, void * );
typedef int (*bpo_Work_List2_t) ( bpo_List_t *, void *, void * );

typedef int (* bpo_sort_i_t)(const void *, const void *);

typedef int (*p_func_t)(FILE *,const char *,...);

/* ------------------------------------------------------------------
    bpo_queue structures
   ------------------------------------------------------------------ */

/* The following is for convenience only. On no account should
   user programs examine or modify this structure in any way.
   Hell, for all you know, I've changed the structure internally.
   Where would that get you ?
*/

struct bpo_Q
{
bpo_Node_t	Q_Node;
bpo_List_t	Q_List;
int		Q_id;
int		Q_Count;
int		Q_max_count;
int		Q_notify;
int		Q_flags;

pid_t		Q_sole_pid;

bpo_List_t	Q_pid_r_list;

char		Q_name[1];
};

struct bpo_Q_Mesg
{
bpo_Node_t	Q_Node;		/* Node for linking to Q */
off_t		reply_Q;	/* offset to reply Q */
int		reply_id;	/* id of reply Q (sanity check) */
};


/* ------------------------------------------------------------------
    function definitions
   ------------------------------------------------------------------ */

extern void *
Create_SODB_mmap( const char *use_name, const char *mmap_name,
		size_t msize, off_t moff, int flags );
		
extern void *
Create_SODB_shm( const char *use_name, const char *fname,
		char id, size_t msize, int flags );

extern void *
Create_SODB_mmap_priv( const char *use_name,
			const char *mmap_name,
			size_t msize, off_t moff,
			int flags, mode_t priv );

extern void *
Create_SODB_shm_priv( const char *use_name,
			const char *fname,
			char id, size_t msize,
			int flags, mode_t priv );

extern int Detach_SODB_mmap( const void *hint );
extern int Detach_SODB_shm( const void *hint );
extern int Detach_SODB_lump( const void *hint );

extern void *Get_SODB_Pointer( const void *hint, off_t item_offset );
extern off_t Get_SODB_Offset( const void *vp );
extern mode_t Get_SODB_Privs( const void *vp );

extern void Do_SODB_Init( void );
extern void * Get_SODB_Base( const void *hint );
extern int SODB_in_Segment( const void *hint, const void *chk );


/* And the resource functions */
extern off_t
Add_SODB_Resource( void *object, const char *res_name );

extern void *
Find_SODB_Resource( const void *hint, const char *res_name );

extern void *
Fetch_SODB_Resource( const void *hint,
		    const char *name,
		    size_t s_size,
		    Fetch_SODB_Resource_t s_init );


/* and a resource list function */
extern int
Validate_Resources( const void * hint,
		    int p_which_procs,
		    Validate_Resource_t disp,
		    void *param1 );

/* For debugging use only, you have been warned ! */
extern void *
Attach_SODB_image( const char *fname );
extern void *
Attach_SODB_image_raw( const char *fname,
			 off_t foff,
			 size_t fsize );


/* ------------------------------------------------------------------
    bpo memory allocation function definitions
   ------------------------------------------------------------------ */

extern void *shalloc( const void *hint, size_t size );
extern void shfree( void *ptr );
extern char *bpo_strdup( const void *hint, const char *s1 );
extern off_t bpo_strdup_offset( const void *hint, const char *s1 );


/* ------------------------------------------------------------------
    bpo list handling function definitions
   ------------------------------------------------------------------ */

extern bpo_Node_t *
bpo_Init_Node( bpo_Node_t *node, const char *name, void *item );

extern bpo_List_t *
bpo_Init_List( bpo_List_t *list, int flags );

extern bpo_Node_t *
bpo_Remove_Node( bpo_Node_t *node1 );

extern bpo_Node_t *
bpo_Remove_Head( bpo_List_t *list );

extern bpo_Node_t *
bpo_Remove_Tail( bpo_List_t *list );

extern bpo_Node_t *
bpo_Add_Node_After( bpo_Node_t *node1, bpo_Node_t *node2 );

extern bpo_Node_t *
bpo_Add_Node_Before( bpo_Node_t *node1, bpo_Node_t *node2 );

extern bpo_Node_t *
bpo_Add_Head( bpo_List_t *list, bpo_Node_t *node );

extern bpo_Node_t *
bpo_Add_Tail( bpo_List_t *list, bpo_Node_t *node );

extern void *
bpo_Find_Item_By_Name( bpo_List_t *list, const char *name );

extern void *
bpo_Find_Next_Item_By_Name( bpo_Node_t *node, const char *name );

extern bpo_Node_t *
bpo_Find_Node_By_Item( bpo_List_t *list, void *item );

extern void *
bpo_Find_Item_From_Node( bpo_Node_t *node );

extern void *
bpo_Remove_Head_Item( bpo_List_t *list );

extern void *
bpo_Remove_Tail_Item( bpo_List_t *list );

extern void
bpo_Walk_List( bpo_List_t *l_ptr,
		bpo_Walk_List_t disp,
		void *param );

extern int
bpo_Walk_List2( bpo_List_t *l_ptr,
		bpo_Walk_List2_t disp,
		void *param );

extern int
bpo_Walk_List_N( bpo_List_t *l_ptr,
		    bpo_Walk_List_N_t disp,
		    void *param );

extern int
bpo_Work_List( bpo_List_t *list,
		bpo_Work_List_t work,
		void *param );

extern int
bpo_Work_List2( bpo_List_t *list,
		bpo_Work_List2_t work,
		void *param1, void *param2 );

extern int
bpo_May_Work_List( bpo_List_t *list,
		    unsigned long *last_ser_nbr,
		    bpo_Work_List_t work,
		    void *param );

extern int
bpo_May_Work_List2( bpo_List_t *list,
		    unsigned long *last_ser_nbr,
		    bpo_Work_List2_t work,
		    void *param1, void *param2 );

extern unsigned long bpo_Touch_List( bpo_List_t *list );

extern int bpo_Any_In_List( bpo_List_t *list );

extern int
bpo_Transfer_Lists( bpo_List_t *fl_ptr, bpo_List_t *tl_ptr );


extern int
bpo_free_list_contents( bpo_List_t *blp );

extern void
bpo_Free_Node( bpo_Node_t *bnp, int flags );
extern void
bpo_Free_List( bpo_List_t *blp, int flags );


/* The following four functions are especially for the SODB initialisation code
   The object is being created and is not yet in the list of shared objects
   and so the base cannot be looked up and must be passed in explicitly

   Also, the Add_ functions assume that the nodes being added
   are new & clean
*/

extern bpo_Node_t *
bpo_Init_Node_Raw( bpo_Node_t *node,
		    const void *base,
		    const char *name,
		    const void *item );

extern bpo_List_t *
bpo_Init_List_Raw( bpo_List_t *list, const void *base, int flags );

extern bpo_Node_t *
bpo_Add_Head_Raw( bpo_List_t *list, const void *base, bpo_Node_t *node );

extern bpo_Node_t *
bpo_Add_Tail_Raw( bpo_List_t *list, const void *base, bpo_Node_t *node );


/* And a couple of sort functions. */

extern int
bpo_sort_i( bpo_List_t *sort_list, bpo_sort_i_t compar );
extern int
bpo_sort_n( bpo_List_t *sort_list );


/* And for the suspicious */
extern int bpo_Validate_List( bpo_List_t *rlp );
extern int bpo_List_Chk( bpo_List_t *rlp,
			p_func_t p_func,
			FILE *fh  );

extern int node_list_header( p_func_t p_func, FILE *fh );
extern int node_list_details( void *base, bpo_Node_t *bnp,
				p_func_t p_func,
				FILE *fh );


/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/* ------------------------------------------------------------------
    function definitions
   ------------------------------------------------------------------ */

extern int bpo_nap_a_while( long secs, long usecs );

extern int bpo_take_a_nap( void );

extern long b_Lockval( b_lock_t *lockp );

extern void b_Init( b_lock_t *lockp );

extern int b_TryLock( b_lock_t *lockp );

extern void b_Lock( b_lock_t *lockp );

extern void b_Unlock( b_lock_t *lockp );

/* The following seven functions are now deprecated
    and may be removed in a later release of library
*/
extern void bpo_RLock( struct bpo_fl_lock *bflp );
extern void bpo_WLock( struct bpo_fl_lock *bflp);
extern void bpo_RUnlock( struct bpo_fl_lock *bflp );
extern void bpo_WUnlock( struct bpo_fl_lock *bflp );
extern int bpo_RWLock( struct bpo_fl_lock *bflp );
extern void bpo_WRLock( struct bpo_fl_lock *bflp );
extern void bpo_Init_RWLock( struct bpo_fl_lock *bflp );



extern struct bpo_pid_lock *
	bpo_Init_pid_lock( struct bpo_pid_lock *bplp,
			    struct timeval *udtp );

extern void bpo_pid_Lock( struct bpo_pid_lock *bplp );

extern void bpo_pid_Unlock( struct bpo_pid_lock *bplp );

extern int bpo_pid_set_flags( struct bpo_pid_lock *bplp, int set_bits );
extern int bpo_pid_clr_flags( struct bpo_pid_lock *bplp, int clr_bits );
extern int bpo_pid_get_flags( struct bpo_pid_lock *bplp );


/* ------------------------------------------------------------------
    bpo function definitions
   ------------------------------------------------------------------ */

/* functions that deal exclusively with ourself */
extern int
Set_My_Proc_Details( const char *group_name,
			const char *prog_name,
			const char *fifo_prefix,
			const char *fifo_name,
			mode_t fifo_mode,
			int flags );

#ifndef MPLIB1_BPO_INTERNAL
extern void * Get_My_Proc_Details( mode_t privs );
extern void * Check_My_Proc_Details( void );
#endif

extern ssize_t Wait_Process_Pipe( void );
extern int Get_Process_Pipe_fd( void );

/* Now related to a given segment */
extern int Remove_Shm_Process( const void *hint, const char *pid_str );
extern int Register_Process_Details( const void *hint );
extern int Deregister_Process_Details( const void *hint );

extern int Add_This_Pid_Resource( const void *hint, bpo_Node_t *nptr );

extern ssize_t Inform_Named_Pipe( const char *pipe_name );
extern ssize_t Inform_Process_Str( const void *hint, const char *pid_str );
extern ssize_t Inform_Process_Pid( const void *hint, pid_t pid );

extern int scan_dead_pipes( const char *dir_name );

/* And the higher level standard comms attached routine */
extern void *get_comms_hint( void );
extern void *comms_shm_attach( const char *config_name, int flags );
extern void comms_shm_detach( void );





/* ------------------------------------------------------------------
    function definitions
   ------------------------------------------------------------------ */

extern int mpPutMsg( const char *q_name, bpo_Node_t *q_node );
extern void *mpGetMsg( const char *q_name, const void *hint );

extern int mpReplyMsg( struct bpo_Q_Mesg *mesp );
extern int mpPostMsg( const char *q_name,
		    struct bpo_Q_Mesg *mesp,
		    struct bpo_Q *reply_q );

extern void * mpFindQ( const char *q_name, const void *hint );
extern int mpCheckQ( const char *q_name, const void *hint );
extern int mpWaitQ( const char *name, const void *hint );

extern struct bpo_Q *
mpCreateQ( const void *hint, const char *q_name,
	    int q_max, int q_flags, int q_mech );

extern void *
bpo_send_and_get( const char *q_name,
		struct bpo_Q_Mesg *bpo_msg,
		struct bpo_Q *reply_q );


/* ------------------------------------------------------------------
    daemon process function definition
   ------------------------------------------------------------------ */

extern int
bpo_daemon_sleep( int how_long );

#endif
/* -- End of File -- */
