/*
*******************************************************************************
*  Copyright (c) 1997 Martin Poole                            *
*******************************************************************************
**
*******************************************************************************
*/

/* ------------------------------------------------------------------
    Include files
   ------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <signal.h>

#include <mplib1/dl_list.h>
#include <mplib1/pid_check.h>
#include <mplib1/cfg_file.h>
#include "include/bpo_internal.h"
#include <mplib1/bpo.h>
#include <mplib1/daemon.h>
#include <mplib1/fprintfile.h>

/* ------------------------------------------------------------------
    defines
   ------------------------------------------------------------------ */

/*
Lock test code
        set stderr.

        :restart_point
        fork()
        parent dies.
        if parent_pid set, wait for parent to die.
        register self.
        set parent_pid = getpid()
        clean process.
        clean queues.
        Check for bad queue list, die if bad.
        Create 'n' queues (some Q_NM_NONE, some Q_NM_PIPE)
        Craete 'n' messages.
        Post messages to queues.
        goto restart_point.

*/


struct l_test
{
struct	bpo_pid_lock	l_test1;
};

static char lt_str[]="LOCK_TEST";

/* ------------------------------------------------------------------
    Code starts here
   ------------------------------------------------------------------ */

static void
lt_init( struct l_test *ltptr )
{
    bpo_Init_pid_lock( &ltptr->l_test1, NULL );
    return;
}

static void
do_lock_test( void *hint )
{
    struct l_test *ltptr;
    int t;
    ltptr = Fetch_SODB_Resource( hint, lt_str,
				sizeof(struct l_test),
				(Fetch_SODB_Resource_t)lt_init );
    if (ltptr)
    {
	/* First we lock the structure */
	fprintfile(stderr, "Parent: about to lock \n" );
	bpo_pid_Lock( &ltptr->l_test1 );
	fprintfile(stderr, "Parent: about to fork \n" );
	t = fork();
	if (t==0) /* We are child */
	{
	    fprintfile(stderr, "Child: about to attempt lock\n" );
	    bpo_pid_Lock( &ltptr->l_test1 );
	    fprintfile(stderr, "Child acquired lock\n" );
	    bpo_pid_Unlock( &ltptr->l_test1 );
	}else
	{
	    /* assume success */
	    fprintfile(stderr, "Parent: about to snooze\n" );
	    sleep(1);
	    fprintfile(stderr, "Parent: about to exit\n" );
	    exit(0);
	}
    }else
    {
	fprintfile(stderr, "Unable to allocate test resoruce\n" );
    }
    return;
}

int
main( int argc, char *argv[] )
{
    void *hint;
    char *shm_name,*e_nm=NULL;
    char *cp=NULL;
    int optc;

/*    set_config_string( "STDERR", "q_test.err" );*/
    while ((optc = getopt(argc, argv, "e:f:")) != EOF)
    switch(optc)
    {
    case 'e' : /* Specify config file */
	e_nm =optarg;
	break;
    case 'f' : /* Specify config file */
	read_config_file(cp=optarg);
	break;
    default:
/*	usage( argv[0], "Invalid argument supplied" );*/
	break;
    }
    if (cp==NULL)
    {
	cp = eval_config_default( NULL, "$HOME/config/q_test.cfg" );
	read_config_file( cp );
    }
    if (e_nm)
    {
	set_config_string( "STDERR", e_nm );
    }

    shm_name = strdup(eval_config_default( "SHM_CONFIG", "$HOME/config/shm.cfg" ));
    hint = comms_shm_attach( shm_name, SODB_READ_CONFIG );
    fprintfile(stderr,"Attach returned: %p\n", hint );
    if (hint)
    {
	do_lock_test( hint );
    }
    return(0);
}

/* -- End of File -- */
